from typing import Optional
import numpy as np
import numpy.typing as npt
from . import rendered
from .loss import compute as compute_loss, Loss


def solve(
    k: npt.ArrayLike,
    y: npt.ArrayLike,
    loss: Loss,
    lmbda: float,
    z0: Optional[tuple[npt.ArrayLike, float]] = None,
    tol: float = 1e-16,
    max_steps: int = 100,
    max_back_steps: int = 10,
    sigma: float = 0.01,
    eta: float = 0.1,
    verbose: int = 0,
    regularization: float = 1e-10,
):
    # initialize
    k = np.array(k)
    n = k.shape[0]
    if z0 is None:
        a = np.zeros(n)
        b = loss.mean2param(np.mean(y))
        ka = np.zeros(n)
    else:
        a = np.array(z0[0])
        b = z0[1]
        ka = k.dot(a) / lmbda
    t = ka + b

    for step in range(max_steps):
        # compute objective
        reg = 0.5 * ka.dot(a)
        ell, g, h = compute_loss(y, t, loss, derivative=True)
        obj = reg + ell.mean()

        if np.isinf(obj) or np.isnan(obj):
            raise ValueError('objective is inf or nan')
        g /= n
        h /= n

        # compute violation
        rhs_a = a + g
        rhs_b = a.sum()
        rhssqr = (lmbda * rhs_a.dot(rhs_a) + rhs_b**2) / (n + 1)
        # output
        if verbose >= 1:
            print(step + 1, obj, rhssqr, end='')
        # check for termination
        if rhssqr < tol:
            if verbose >= 1:
                print(
                    f'\noptimization terminated after {step}', 'steps (violation < tol)'
                )
            break

        # solve Newton system
        rhs = (rhs_a, rhs_b)
        if h.sum() < regularization:
            h += regularization
        d_a, d_b = rendered.solve(k, h, lmbda, rhs)
        dec = k.dot(rhs_a).dot(d_a) / lmbda + g.sum() * d_b
        if verbose >= 1:
            print('', dec)

        an, bn, kan, tn = a, b, ka, t
        s = 1.0
        for _backstep in range(max_back_steps):
            # update vars
            an = a - s * d_a
            bn = b - s * d_b
            # evaluate
            kan = k.dot(an) / lmbda
            tn = kan + bn
            regn = 0.5 * kan.dot(an)
            elln = compute_loss(y, tn, loss)

            objn = regn + elln.mean()
            # output
            if verbose >= 2:
                print(' > ', _backstep + 1, s, objn - obj)
            # check step
            if objn < obj - sigma * s * dec + tol:
                break
            # decrease step size
            s *= eta

        if (obj - objn) * lmbda < tol:
            if verbose >= 1:
                print(
                    f'\noptimization terminated after {step+1}',
                    'steps (decrease < tol)',
                )
            break

        # update
        a, b, ka, t = an, bn, kan, tn
    else:
        if verbose >= 1:
            print('optimization did not terminate')
    return a, b



import numpy as np
import matplotlib.pyplot as plt
import gkm


def fit_and_plot(x, y, lmbda, gamma, loss, ax):
    xsqr = (x ** 2).sum(axis=1)
    dsqr = xsqr[:, None] + xsqr[None, :] - 2.0 * x.dot(x.T)
    k = np.exp(-gamma * dsqr)

    clf = gkm.GKM(lmbda=lmbda, loss=loss)
    clf.fit_with_kernel_matrix(k, y)

    yp = clf.predict_with_kernel_matrix(k)
    ax.scatter(x[:, 0], x[:, 1], s=np.maximum(0.0, yp))


def main():
    # generate data
    np.random.seed(5)
    n, nft = 500, 3
    x = np.random.rand(n, nft)
    y = 50.0 * np.prod(np.sin(2.0 * np.pi * x[:, :2]), axis=1) ** 2

    # rescale features
    xsqr = (x ** 2).sum(axis=1)
    dsqr = xsqr[:, None] + xsqr[None, :] - 2.0 * x.dot(x.T)
    mscale = np.mean(dsqr)
    x /= np.sqrt(mscale)

    # optimize
    loss = gkm.deviance.poisson
    clf = gkm.OptimizedGKM(
        loss=loss,
        regularization=gkm.regularization.penalize_gamma(1e-6),
        lmbda0=1e-3,
        gamma0='auto',
        kernel_type='multi-rbf',
        verbose=1,
    )
    clf.fit(x, y)

    print('lmbda:', clf.lmbda0, '->', clf.lmbda)
    print('gamma:', clf.gamma0[0], '->', clf.gamma[0])
    print('value:', clf.value0, '->', clf.value)

    # visualize resulting decision functions
    fig, ax = plt.subplots(1, 3, figsize=(10, 4))
    fit_and_plot(x, y, clf.lmbda0, clf.gamma0[0], loss, ax[0])
    ax[0].set_title('initial predictions')
    fit_and_plot(x, y, clf.lmbda, clf.gamma[0], loss, ax[1])
    ax[1].set_title('final predictions')
    ax[2].scatter(x[:, 0], x[:, 1], s=y)
    ax[2].set_title('true labels')
    plt.show()


if __name__ == '__main__':
    main()



import numpy as np


def penalize_gamma(factor):
    def _regularization(data):
        gamma = np.exp(data['logp'][1:])
        gamma0 = np.exp(data['logp0'][1:])
        dim = gamma.size
        reg = gamma / gamma0
        dreg = np.concatenate(([0.0], reg / dim))
        return factor * reg.mean(), factor * dreg

    return _regularization



import numpy as np
from scipy.special import xlogy
from .base import Deviance


class Binomial(Deviance):
    def __init__(self, n=1):
        super().__init__()
        self.n = n

    def log_density(self, y, t):
        return y * t - self.n * np.log(1.0 + np.exp(t))

    def __call__(self, y, t):
        ell = -self.log_density(y, t)
        ell -= xlogy(self.n, self.n) - xlogy(y, y) - xlogy(self.n - y, self.n - y)
        return ell

    def param2mean(self, t):
        return self.n / (1.0 + np.exp(-t))

    def dparam2mean(self, t):
        ent = np.exp(-t)
        return self.n / (1.0 + ent) ** 2 * ent

    def mean2param(self, mu):
        return -np.log(self.n / mu - 1.0)

    def dmean2param(self, mu):
        return 1.0 / mu + 1.0 / (self.n - mu)



from .base import Deviance
from ..helpers import safe_log


class Exponential(Deviance):
    def log_density(self, y, t):
        return y * t + safe_log(-t)

    def __call__(self, y, t):
        return -self.log_density(y, t) - 1 - safe_log(y)

    def param2mean(self, t):
        return -1.0 / t

    def dparam2mean(self, t):
        return 1.0 / t**2

    def mean2param(self, mu):
        return -1.0 / mu

    def dmean2param(self, mu):
        return 1.0 / mu**2



import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from . import deviance
from .loss import compute as Loss


class GLM(BaseEstimator, RegressorMixin):
    def __init__(
        self,
        loss: Loss = deviance.gaussian,
        verbose: int = 0,
        tol: float = 1e-16,
        max_steps: int = 1000,
        add_offset: bool = False,
    ):
        super().__init__()
        self.verbose = verbose
        self.loss = loss
        self.tol = tol
        self.add_offset = add_offset
        self.max_steps = max_steps

    def fit(self, X, y, groups=None):
        X, y = check_X_y(X, y)
        n, nft = X.shape
        assert y.shape == (n,)
        loss = self.loss

        if self.add_offset:
            raise NotImplementedError()

        offset = loss.mean2param(np.mean(y))
        beta = np.linalg.lstsq(X, offset * np.ones(n), rcond=None)[0]
        t = X.dot(beta)

        for step in range(self.max_steps):
            ell = loss(y, t).sum()
            dell = loss.deriv(y, t)
            dell_sqr = dell.dot(dell)
            if self.verbose >= 1:
                print(step + 1, ell, dell_sqr)
            if dell_sqr < self.tol:
                if self.verbose >= 1:
                    print(
                        f'\noptimization terminated after {step}',
                        'steps (violation < tol)',
                    )
                break
            ddell = loss.deriv2(y, t)
            # dbeta = np.linalg.lstsq(ddell[:, None] * X, dell, rcond=None)[0]
            dbeta = np.linalg.solve(X.T.dot(ddell[:, None] * X), X.T.dot(dell))
            # print(dell, ddell)
            print(X.T.dot(ddell[:, None] * X))
            tau = 1.0
            for step_back in range(10):
                betan = beta - tau * dbeta
                tn = X.dot(betan)
                elln = loss(y, tn).sum()
                if self.verbose >= 2:
                    print(' ', step_back, elln - ell)
                if elln < ell:
                    break
                tau *= 0.1
            if ell - elln < self.tol:
                if self.verbose >= 1:
                    print(
                        f'\noptimization terminated after {step+1}',
                        'steps (decrease < tol)',
                    )
                break
            beta = betan
            t = tn

        self._beta = beta

    @property
    def coeffs(self):
        return self._beta

    def predict(self, X):
        check_is_fitted(self, ['_beta'])
        X = check_array(X)
        t = X.dot(self._beta)
        mu = self.loss.param2mean(t)
        return mu

    def sensitivity_prediction(self, mu, x):
        t = x.dot(self._beta)
        dmu = 1.0 / self.loss.dmean2param(mu)
        return dmu * t



import numpy as np
from numpy.testing import assert_allclose
import gkm


def test_gaussian_deriv():
    tau = 1e-8
    np.random.seed(1)

    n = 1000
    y = np.random.randn(n)

    loss = gkm.deviance.gaussian
    t = loss.mean2param(np.random.randn(n))

    l0 = loss(y, t)
    dl0 = loss.deriv(y, t)
    ddl0 = loss.deriv2(y, t)

    l1 = loss(y, t + tau)
    dl_approx = (l1 - l0) / tau
    assert_allclose(dl0, dl_approx, rtol=1e-5, atol=1e-12)

    dl1 = loss.deriv(y, t + tau)
    ddl_approx = (dl1 - dl0) / tau
    assert_allclose(ddl0, ddl_approx, rtol=1e-5, atol=1e-12)


def test_poisson_deriv():
    tau = 1e-8
    np.random.seed(1)

    n = 1000
    y = np.random.poisson(size=n)

    loss = gkm.deviance.poisson
    t = loss.mean2param(1.0 + np.random.poisson(size=n))

    l0 = loss(y, t)
    dl0 = loss.deriv(y, t)
    ddl = loss.deriv2(y, t)

    l1 = loss(y, t + tau)
    dl_approx = (l1 - l0) / tau
    assert_allclose(dl0, dl_approx, rtol=1e-5, atol=1e-6)

    dl1 = loss.deriv(y, t + tau)
    ddl_approx = (dl1 - dl0) / tau
    assert_allclose(ddl, ddl_approx, rtol=1e-5, atol=1e-6)


def test_binomial_deriv():
    tau = 1e-8
    np.random.seed(1)

    n = 1000
    y = np.random.binomial(5, 0.5, n)
    mu = 1.0 / (1.0 + np.exp(np.random.randn(n)))

    loss = gkm.deviance.Binomial(5)

    l0 = loss(y, mu)
    dl0 = loss.deriv(y, mu)
    ddl = loss.deriv2(y, mu)

    l1 = loss(y, mu + tau)
    dl_approx = (l1 - l0) / tau
    assert_allclose(dl0, dl_approx, rtol=1e-4, atol=1e-12)

    dl1 = loss.deriv(y, mu + tau)
    ddl_approx = (dl1 - dl0) / tau
    assert_allclose(ddl, ddl_approx, rtol=1e-4, atol=1e-12)


def test_geometric_deriv():
    tau = 1e-8
    np.random.seed(1)

    n = 1000
    y = np.random.geometric(0.5, n)

    loss = gkm.deviance.geometric
    t = loss.mean2param(np.random.geometric(0.5, n))

    l0 = loss(y, t)
    dl0 = loss.deriv(y, t)
    ddl = loss.deriv2(y, t)

    l1 = loss(y, t + tau)
    dl_approx = (l1 - l0) / tau
    assert_allclose(dl0, dl_approx, rtol=1e-4, atol=1e-6)

    dl1 = loss.deriv(y, t + tau)
    ddl_approx = (dl1 - dl0) / tau
    assert_allclose(ddl, ddl_approx, rtol=1e-4, atol=1e-6)



import numpy as np
from .multi_rbf import MultiRBF


class MultiRBFLazy(MultiRBF):
    def _estimate_gamma(self):
        return np.log(2.0) / np.maximum(
            [
                ((self._x[:, i][None, :] - self._x[:, i][:, None]) ** 2).mean()
                for i in range(self.dim)
            ],
            1e-10,
        )

    def prepare(self, X):
        self._x = X

    def __call__(self, gamma, idx1=slice(None), idx2=slice(None)):
        assert gamma.shape == (self.dim,)
        xs = self._x * np.sqrt(gamma)[None]
        xssqr = (xs**2).sum(axis=1)
        ndssqr = (
            2.0 * xs[idx1].dot(xs[idx2].T) - xssqr[idx1][:, None] - xssqr[idx2][None, :]
        )
        return np.exp(ndssqr)

    def deriv(self, gamma, k, idx1=slice(None), idx2=slice(None)):
        for i in range(self.dim):
            dsqr = (self._x[idx1, i][:, None] - self._x[idx2, i][None, :]) ** 2
            yield -k[idx1][:, idx2] * dsqr

    def compute(self, gamma, X):
        xs = self._x * np.sqrt(gamma)[None]
        xssqr = (xs**2).sum(axis=1)
        x1s = X * np.sqrt(gamma)[None]
        x1ssqr = (x1s**2).sum(axis=1)
        ndssqr = 2.0 * xs.dot(x1s.T) - xssqr[:, None] - x1ssqr[None, :]
        return np.exp(ndssqr)



import numpy as np
import matplotlib.pyplot as plt
import gkm


def main():
    np.random.seed(5)
    n, nft = 1000, 2
    x = np.random.rand(n, nft)
    y = 50.0 * np.prod(np.sin(2.0 * np.pi * x), axis=1) ** 2
    y = np.floor(y)

    xsqr = (x ** 2).sum(axis=1)
    dsqr = xsqr[:, None] + xsqr[None, :] - 2.0 * x.dot(x.T)
    mscale = np.mean(dsqr)
    x /= np.sqrt(mscale)

    clf = gkm.GKM(lmbda=1e-6, loss=gkm.deviance.Geometric(), verbose=2)
    clf.fit(x, y)

    f = clf.predict(x)
    print(np.hstack((y[:, None], np.floor(f[:, None]))))

    fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    ax[0].scatter(x[:, 0], x[:, 1], s=y)
    ax[1].scatter(x[:, 0], x[:, 1], s=np.maximum(0.0, f))
    plt.show()


if __name__ == '__main__':
    main()



from .gkm import GKM
from .glm import GLM
from .optimized_gkm import OptimizedGKM
from . import deviance
from . import links
from . import regularization
from . import validation
from .loss import Loss

__all__ = [
    'GKM',
    'GLM',
    'OptimizedGKM',
    'deviance',
    'links',
    'regularization',
    'validation',
    'Loss',
]



import itertools
import pytest
from numpy.testing import assert_allclose
import numpy as np
import gkm
from common import get_data, compute_kernel, idfn


@pytest.mark.parametrize('lmbda,loss', itertools.product(
    [1.0, 0.1, 1e-3],
    [gkm.deviance.gaussian, gkm.deviance.poisson, gkm.deviance.geometric],
), ids=idfn)
def test_sensitivity(lmbda, loss):
    tau = 1e-6
    np.random.seed(5)
    n, nft = 10, 2
    x, y = get_data(n, nft)
    nv = 5
    xv, yv = get_data(nv, nft)

    loglmbda = np.log(lmbda)
    k = compute_kernel(x, x)
    clf = gkm.GKM(lmbda=np.exp(loglmbda), loss=loss)
    clf.fit_with_kernel_matrix(k, y)
    s = clf.sensitivity_solution()

    kv = compute_kernel(xv, x)
    fv = clf.predict_with_kernel_matrix(kv).sum()
    dfv = np.ones(nv)

    t = kv.dot(clf._a) / lmbda + clf._b
    dmu = loss.dparam2mean(t)
    dfv *= dmu

    # sensitivity for lmbda
    clf_n = gkm.GKM(lmbda=np.exp(loglmbda + tau), loss=loss)
    clf_n.fit_with_kernel_matrix(k, y)
    kv_n = compute_kernel(xv, x)
    fv_n = clf_n.predict_with_kernel_matrix(kv_n).sum()
    dfv_approx = (fv_n.sum() - fv.sum()) / tau
    da, db = s['loglmbda']
    assert_allclose(
        dfv.dot(kv.dot(da - clf._a) / lmbda + db),
        dfv_approx,
        rtol=1e-4, atol=1e-6,
    )

    # sensitivity for kernel (train)
    clf_n = gkm.GKM(lmbda=np.exp(loglmbda), loss=loss)
    clf_n.fit_with_kernel_matrix(k ** np.exp(tau), y)
    kv_n = compute_kernel(xv, x)
    fv_n = clf_n.predict_with_kernel_matrix(kv_n).sum()
    dfv_approx = (fv_n.sum() - fv.sum()) / tau
    da, db = s['kernel'](k * np.log(k))
    assert_allclose(
        dfv.dot(kv.dot(da) / lmbda + db),
        dfv_approx,
        rtol=1e-4, atol=1e-6,
    )

    # sensitivity for kernel (validate)
    clf_n = gkm.GKM(lmbda=np.exp(loglmbda), loss=loss)
    clf_n.fit_with_kernel_matrix(k, y)
    kv_n = compute_kernel(xv, x) ** np.exp(tau)
    fv_n = clf_n.predict_with_kernel_matrix(kv_n).sum()
    dfv_approx = (fv_n.sum() - fv.sum()) / tau
    da, db = s['kernel'](np.zeros_like(k))
    assert_allclose(
        dfv.dot((kv.dot(da) + (kv * np.log(kv)).dot(clf._a)) / lmbda + db),
        dfv_approx,
        rtol=1e-4, atol=1e-6,
    )

    # sensitivity for kernel (both)
    clf_n = gkm.GKM(lmbda=np.exp(loglmbda), loss=loss)
    clf_n.fit_with_kernel_matrix(k ** np.exp(tau), y)
    kv_n = compute_kernel(xv, x) ** np.exp(tau)
    fv_n = clf_n.predict_with_kernel_matrix(kv_n).sum()
    dfv_approx = (fv_n.sum() - fv.sum()) / tau
    da, db = s['kernel'](k * np.log(k))
    assert_allclose(
        dfv.dot((kv.dot(da) + (kv * np.log(kv)).dot(clf._a)) / lmbda + db),
        dfv_approx,
        rtol=1e-4, atol=1e-6,
    )



import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from . import newton
from . import rendered
from . import deviance
from . import validation
from .loss import compute as compute_loss, Loss
from .kernel import get_kernel


class GKM(BaseEstimator, RegressorMixin):
    def __init__(
        self,
        lmbda: float,
        loss: Loss = deviance.gaussian,
        verbose: int = 0,
        tol: float = 1e-16,
        kernel_type: str = 'rbf',
        gamma: float = 1.0,
    ):
        super().__init__()
        self.lmbda = lmbda
        self.verbose = verbose
        self.loss = loss
        self.tol = tol
        self.kernel_type = kernel_type
        self.kernel = get_kernel(kernel_type)
        self.gamma = self.kernel.get_initial(gamma)

    def fit(self, X, y, groups=None):
        X, y = check_X_y(X, y)
        self.kernel.prepare(X)
        k = self.kernel(self.gamma)
        return self.fit_with_kernel_matrix(k, y)

    def fit_with_kernel_matrix(self, k, y):
        # solve training problem using Newton's method
        self._k = k
        self._y = y
        a, b = newton.solve(
            k,
            y,
            self.loss,
            self.lmbda,
            verbose=self.verbose,
            tol=self.tol,
        )
        self._a = a
        self._b = b
        return self

    def predict(self, X):
        check_is_fitted(self, ['_a', '_b'])
        X = check_array(X)
        k = self.kernel.compute(self.gamma, X)
        return self.predict_with_kernel_matrix(k)

    def predict_with_kernel_matrix(self, k):
        check_is_fitted(self, ['_k', '_y', '_a', '_b'])
        ka = k.dot(self._a)
        t = ka / self.lmbda + self._b
        mu = self.loss.param2mean(t)
        return mu

    def sensitivity_solution(self):
        ka = self._k.dot(self._a)
        t = ka / self.lmbda + self._b
        h = compute_loss(self._y, t, self.loss, derivative=True)[2]
        h /= h.size
        return dict(
            loglmbda=rendered.solve(
                self._k,
                h,
                self.lmbda,
                (
                    h * ka / self.lmbda,
                    0.0,
                ),
            ),
            kernel=lambda dk: rendered.solve(
                self._k,
                h,
                self.lmbda,
                (
                    -h * np.tensordot(dk, self._a, axes=(1, 0)) / self.lmbda,
                    0.0,
                ),
            ),
        )

    def sensitivity_prediction(self, mu, dk):
        dmu = 1.0 / self.loss.dmean2param(mu)
        return dmu[:, None] * np.tensordot(dk, self._a, axes=(1, 0))

    def sensitivity(self, kv, de):
        ka = self._k.dot(self._a)
        t = ka / self.lmbda + self._b
        h = compute_loss(self._y, t, self.loss, derivative=True)[2]
        h /= h.size
        if np.isnan(h).any():
            raise FloatingPointError()

        kva = kv.dot(self._a)
        tv = kva / self.lmbda + self._b
        dmuv = self.loss.dparam2mean(tv)
        if np.isnan(dmuv).any():
            raise FloatingPointError()
        gamma = rendered.solve_transposed(
            self._k,
            h,
            self.lmbda,
            (
                (de * dmuv).dot(kv) / self.lmbda,
                (de * dmuv).sum(),
            ),
        )[0]

        def dkernel(dk, dkv):
            return (
                np.tensordot(
                    (
                        np.tensordot(dkv, de * dmuv, axes=(1, 0))
                        - np.tensordot(h[None, None, :] * dk, gamma, axes=(2, 0))
                    ),
                    self._a,
                    axes=(1, 0),
                )
                / self.lmbda
            )

        return dict(
            loglmbda=(gamma.dot(h * ka) - (de * dmuv).dot(kva)) / self.lmbda,
            kernel=dkernel,
        )

    def score(self, x, y):
        y_est = self.predict(x)
        t_est = self.loss.mean2param(y_est)
        t_const = self.loss.mean2param(np.mean(self._y))
        return self.loss(y, t_est).sum() / self.loss(y, t_const * np.ones_like(y)).sum()

    def objective(self, x, y, groups=None, cv=None):
        self.kernel.prepare(x)
        k = self.kernel(self.gamma)
        return validation.score(
            self,
            self.gamma,
            k,
            y,
            groups,
            cv=cv,
            kernel=self.kernel,
            loss=self.loss,
        )



import numpy as np
from numpy.testing import assert_allclose
from gkm import rendered


def test_rendered_solve_positive():
    np.random.seed(5)
    n = 10
    for trial in range(10):
        kh = np.random.randn(n, n // 2)
        k = kh.dot(kh.T)
        h = np.random.randn(n) ** 2
        # h = np.concatenate((np.random.randn(n // 2) ** 2, np.zeros(n // 2)))
        lmbda = 1e-3
        rhs_a = np.random.rand(n)
        rhs_b = np.random.rand()
        rhs = rhs_a, rhs_b

        da, db = rendered.solve_positive(k, h, lmbda, rhs)
        rec_a = da + h * k.dot(da) / lmbda + h * db
        rec_b = da.sum()
        assert_allclose(rec_b, rhs_b)
        assert_allclose(rec_a, rhs_a)

        d = np.linalg.solve(np.block([
            [np.eye(n) + h[:, None] * k / lmbda, h[:, None]],
            [np.ones((1, n)), 0.0],
        ]), np.concatenate([rhs_a, [rhs_b]]))
        assert_allclose(np.concatenate((da, [db])), d)

        da, db = rendered.solve_transposed_positive(k, h, lmbda, rhs)
        rec_a = da + k.dot(h * da) / lmbda + db
        rec_b = np.dot(h, da)
        assert_allclose(rec_b, rhs_b)
        assert_allclose(rec_a, rhs_a)

        d = np.linalg.solve(np.block([
            [np.eye(n) + h[:, None] * k / lmbda, h[:, None]],
            [np.ones((1, n)), 0.0],
        ]).T, np.concatenate([rhs_a, [rhs_b]]))
        assert_allclose(np.concatenate((da, [db])), d)


def test_rendered_solve():
    np.random.seed(5)
    n = 10
    for trial in range(10):
        kh = np.random.randn(n, n // 2)
        k = kh.dot(kh.T)
        h = np.concatenate((np.random.randn(n // 2) ** 2, np.zeros(n // 2)))
        lmbda = 1e-3
        rhs_a = np.random.rand(n)
        rhs_b = np.random.rand()

        da, db = rendered.solve(k, h, lmbda, (rhs_a, rhs_b))

        rec_a = da + h * k.dot(da) / lmbda + h * db
        rec_b = da.sum()
        assert_allclose(rec_b, rhs_b)
        assert_allclose(rec_a, rhs_a)

        d = np.linalg.solve(np.block([
            [np.eye(n) + h[:, None] * k / lmbda, h[:, None]],
            [np.ones((1, n)), 0.0],
        ]), np.concatenate([rhs_a, [rhs_b]]))
        assert_allclose(np.concatenate((da, [db])), d)



import numpy as np
from .rbf import RBF


class MultiRBF(RBF):
    def get_bounds(self, bounds):
        return [bounds] * self.dim

    def _estimate_gamma(self):
        return np.log(2.0) / np.maximum(self._dsqr.mean(axis=(1, 2)), 1e-10)

    def prepare(self, X):
        self._x = X
        self._dsqr = np.maximum((X.T[:, :, None] - X.T[:, None, :]) ** 2, 0.0)

    def __call__(self, gamma, idx1=slice(None), idx2=slice(None)):
        assert gamma.shape == (self.dim,)
        return np.exp(
            np.tensordot(
                self._dsqr[:, idx1][:, :, idx2],
                -gamma,
                axes=(0, 0),
            )
        )

    def deriv(self, gamma, k, idx1=slice(None), idx2=slice(None)):
        return -k[idx1][:, idx2] * self._dsqr[:, idx1][:, :, idx2]

    def compute(self, gamma, X):
        dsqr = np.maximum((X.T[:, :, None] - self._x.T[:, None, :]) ** 2, 0.0)
        return np.exp(np.tensordot(dsqr, -gamma, axes=(0, 0)))

    @property
    def dim(self):
        return self._x.shape[1]



import numpy as np
from scipy.optimize import OptimizeResult


def amsgrad(
    f,
    x0,
    n_iter: int = 100,
    alpha: float = 0.001,
    beta_grad: float = 0.9,
    beta_gradsqr: float = 0.99,
    beta_fun: float = 0.9,
    epsilon: float = 1e-12,
    verbose: int = 0,
    callback=None,
):
    x = np.array(x0)
    n = x0.shape
    avg_grad = np.zeros(n)
    avg_gradsqr = np.zeros(n)
    avg_gradsqr_max = np.zeros(n)
    avg_fun = None

    for t in range(1, n_iter + 1):
        fun, grad = f(x)
        if avg_fun is None:
            avg_fun = fun
        else:
            avg_fun = beta_fun * avg_fun + (1.0 - beta_fun) * fun
        avg_gradsqr = beta_gradsqr * avg_gradsqr + (1.0 - beta_gradsqr) * grad**2
        if verbose > 0:
            print(t, avg_fun, avg_gradsqr.mean())
        if callback is not None:
            try:
                callback(
                    OptimizeResult(
                        x=x,
                        success=True,
                        fun=avg_fun,
                        jac=avg_grad,
                        nit=t,
                    )
                )
            except StopIteration:
                break
        avg_grad = beta_grad**t * avg_grad + (1.0 - beta_grad**t) * grad
        avg_gradsqr_max = np.maximum(avg_gradsqr_max, avg_gradsqr)
        x -= alpha * avg_grad / (np.sqrt(avg_gradsqr_max) + epsilon)
    return OptimizeResult(
        x=x,
        success=True,
        fun=avg_fun,
        jac=avg_grad,
        nit=t,
    )



import numpy as np


class Loss:
    def __call__(self, y, t):
        raise NotImplementedError()

    def deriv(self, y, t):
        raise NotImplementedError()

    def deriv2(self, y, t):
        raise NotImplementedError()

    def param2mean(self, t):
        return t

    def dparam2mean(self, t):
        return np.ones_like(t)

    def mean2param(self, mu):
        return mu

    def dmean2param(self, mu):
        t = self.mean2param(mu)
        return 1.0 / self.dparam2mean(t)


def compute(y, t, loss, derivative=False):
    ell = loss(y, t)
    if not derivative:
        return ell
    dell = loss.deriv(y, t)
    ddell = loss.deriv2(y, t)
    return ell, dell, ddell



import numpy as np
import gkm


def get_data(n, nft):
    # random features
    x = np.random.rand(n, nft)
    # deterministic target
    y = 10.0 * np.prod(np.sin(2.0 * np.pi * x), axis=1) ** 2
    y = np.floor(y)
    # scale features
    xsqr = (x ** 2).sum(axis=1)
    dsqr = xsqr[:, None] + xsqr[None, :] - 2.0 * x.dot(x.T)
    mscale = np.mean(dsqr)
    x /= np.sqrt(mscale) / 2
    return x, y


def compute_kernel(x0, x1=None):
    if x1 is None:
        x1 = x0
    x0sqr = (x0 ** 2).sum(axis=1)
    x1sqr = (x1 ** 2).sum(axis=1)
    return np.exp(2.0 * x0.dot(x1.T) - x0sqr[:, None] - x1sqr[None, :])


def idfn(val):
    if isinstance(val, gkm.Loss):
        return type(val).__name__



import numpy as np
from scipy.special import xlogy
from .base import Deviance
from ..helpers import safe_log


class NegativeBinomial(Deviance):
    def __init__(self, r=1):
        self.r = r

    def log_density(self, y, t):
        return y * t + self.r * safe_log(1.0 - np.exp(t))

    def __call__(self, y, t):
        r = self.r
        ell = -self.log_density(y, t)
        ell += xlogy(r, r) - xlogy(y + r, y + r) + xlogy(y, y)
        return ell

    def param2mean(self, t):
        return self.r * (1.0 / (1.0 - np.exp(t)) - 1.0)

    def dparam2mean(self, t):
        et = np.exp(t)
        return self.r * et / (1.0 - et) ** 2

    def mean2param(self, mu):
        return np.log(mu / (self.r + mu))

    def dmean2param(self, mu):
        return 1.0 / mu - 1.0 / (self.r + mu)



import numpy as np
import matplotlib.pyplot as plt
import gkm


def main():
    np.random.seed(5)
    n, nft = 1000, 2
    x = np.random.rand(n, nft)
    beta = 100 * np.random.rand(nft)
    t = x.dot(beta)
    # sigma = 0.1
    # y = t + sigma * np.random.randn(n)
    y = np.random.poisson(t, size=n)

    clf = gkm.GLM(loss=gkm.deviance.poisson, verbose=2)
    clf.fit(x, y)
    f = clf.predict(x)

    print(beta)
    print(clf.coeffs)

    fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    ax[0].scatter(x[:, 0], x[:, 1], s=y)
    ax[1].scatter(x[:, 0], x[:, 1], s=np.maximum(0.0, f))
    plt.show()


if __name__ == '__main__':
    main()



import numpy as np
from .base import Link


class Identity(Link):
    def __call__(self, mu):
        return mu

    def deriv(self, mu):
        return np.ones_like(mu)

    def deriv2(self, mu):
        return np.zeros_like(mu)

    def inverse(self, t):
        return t

    def inverse_deriv(self, t):
        return np.ones_like(t)

    def inverse_deriv2(self, t):
        return np.zeros_like(t)



from inspect import isgenerator
import numpy as np


def score(clf, gamma, k, y, groups, cv, kernel, loss, deriv=False):
    err = []
    derr = []
    refs = []
    for idx_tr, idx_val in cv.split(k, y, groups):
        kt = k[idx_tr, :][:, idx_tr]
        yt = y[idx_tr]
        clf.fit_with_kernel_matrix(kt, yt)

        kv = k[idx_val, :][:, idx_tr]
        yv = y[idx_val]
        yvp = clf.predict_with_kernel_matrix(kv)
        tvp = loss.mean2param(yvp)
        ttm = loss.mean2param(yt.mean())
        refs.append(loss(yv, ttm * np.ones_like(yvp)).sum())
        err.append(loss(yv, tvp).sum())
        if not deriv:
            continue
        dloss = loss.deriv(yv, tvp) * loss.dmean2param(yvp)
        s = clf.sensitivity(kv, dloss)
        derr_lmbda = s['loglmbda']
        dk_tr = kernel.deriv(gamma, k, idx_tr, idx_tr)
        dk_val = kernel.deriv(gamma, k, idx_val, idx_tr)
        if isgenerator(dk_tr):
            derr_kernel = np.array(
                [
                    s['kernel'](dk_tr_i[None], dk_val_i[None])[0]
                    for dk_tr_i, dk_val_i in zip(dk_tr, dk_val)
                ]
            )
        else:
            derr_kernel = s['kernel'](dk_tr, dk_val)
        derr_kernel *= gamma
        derr.append(np.concatenate(([derr_lmbda], derr_kernel)))
    ref = np.sum(refs)
    avg_err = np.sum(err) / ref
    if not deriv:
        return avg_err
    avg_derr = np.sum(derr, axis=0) / ref
    val = avg_err
    dval = avg_derr
    return val, dval



import numpy as np
from .base import Link
from ..helpers import safe_exp


class Geometric(Link):
    def _clean(self, x):
        return np.clip(1.0 - 1.0 / x, np.finfo(float).eps, np.inf)

    def __call__(self, mu):
        return np.log(self._clean(1.0 + mu))

    def inverse(self, t):
        return 1.0 / (1.0 - safe_exp(t)) - 1.0



import numpy as np
from .base import Link


class Logit(Link):
    def __init__(self, n):
        self.n = n

    def _clean(self, x):
        return np.clip(x, np.finfo(float).eps, self.n - np.finfo(float).eps)

    def __call__(self, mu):
        mu = self._clean(mu)
        return np.log(mu / (self.n - mu))

    def deriv(self, mu):
        mu = self._clean(mu)
        return self.n / (mu * (self.n - mu))

    def deriv2(self, mu):
        mu = self._clean(mu)
        return (2.0 * mu - self.n) / (mu * (self.n - mu)) ** 2

    def inverse(self, t):
        t = np.clip(t, -709.78, 709.78)
        return self.n / (1.0 + np.exp(-t))

    def inverse_deriv(self, t):
        t = np.clip(t, -709.78, 709.78)
        et = np.exp(t)
        return self.n * et / (1.0 + et) ** 2

    def inverse_deriv2(self, t):
        t = np.clip(t, -709.78, 709.78)
        et = np.exp(t)
        return self.n * et * (1.0 - et) / (1.0 + et) ** 2



from .base import Kernel
from .rbf import RBF
from .multi_rbf import MultiRBF
from .multi_rbf_lazy import MultiRBFLazy


def get_kernel(name):
    kernel_cls = {
        'rbf': RBF,
        'multi-rbf': MultiRBF,
        'multi-rbf-lazy': MultiRBFLazy,
    }.get(name, None)
    if kernel_cls is None:
        raise ValueError('unknown kernel')
    return kernel_cls()


__all__ = ['Kernel', 'RBF', 'MultiRBF', 'MultiRBFLazy', 'get_kernel']



import numpy as np


def log_keepnone(v):
    if v is None:
        return v
    return np.log(v)


def log_tuple_keepnone(t):
    return tuple(log_keepnone(v) for v in t)


def safe_exp(x):
    return np.exp(np.clip(x, -709.78, 709.78))


def safe_log(x):
    return np.piecewise(
        x,
        [x > np.finfo(float).eps],
        [np.log, lambda x: -np.inf],
    )



import itertools
import pytest
from numpy.testing import assert_allclose
import numpy as np
import gkm
from common import get_data, compute_kernel, idfn


@pytest.mark.parametrize('lmbda,loss', itertools.product(
    [0.1, 1e-3, 1e-5],
    [gkm.deviance.gaussian, gkm.deviance.poisson, gkm.deviance.geometric],
), ids=idfn)
def test_sensitivity(lmbda, loss):
    tau = 1e-8
    np.random.seed(5)
    n, nft = 5, 2
    x, y = get_data(n, nft)
    nv = 10
    xv, yv = get_data(nv, nft)

    def _validation_error(yvp, deriv=False):
        diff = yvp - yv
        val = (diff ** 2).mean()
        if deriv:
            return val, 2.0 * diff / yv.size
        return val

    loglmbda = np.log(lmbda)
    k = compute_kernel(x, x)
    clf = gkm.GKM(lmbda=np.exp(loglmbda), loss=loss)
    clf.fit_with_kernel_matrix(k, y)
    kv = compute_kernel(xv, x)
    fv, dfv = _validation_error(clf.predict_with_kernel_matrix(kv), deriv=True)
    s = clf.sensitivity(kv, dfv)

    # sensitivity for lmbda
    clf_n = gkm.GKM(lmbda=np.exp(loglmbda + tau), loss=loss)
    clf_n.fit_with_kernel_matrix(k, y)
    fv_n = _validation_error(clf_n.predict_with_kernel_matrix(kv))
    dfv_approx = (fv_n - fv) / tau
    assert_allclose(
        s['loglmbda'],
        dfv_approx,
        rtol=1e-5 / lmbda, atol=1e-6,
    )

    # sensitivity for kernel (train)
    clf_n = gkm.GKM(lmbda=np.exp(loglmbda), loss=loss)
    clf_n.fit_with_kernel_matrix(k ** np.exp(tau), y)
    fv_n = _validation_error(clf_n.predict_with_kernel_matrix(kv))
    dfv_approx = (fv_n - fv) / tau
    assert_allclose(
        s['kernel']((k * np.log(k))[None, :, :], np.zeros_like(kv)[None, :, :]),
        dfv_approx,
        rtol=1e-5 / lmbda, atol=1e-6,
    )

    # sensitivity for kernel (validate)
    clf_n = gkm.GKM(lmbda=np.exp(loglmbda), loss=loss)
    clf_n.fit_with_kernel_matrix(k, y)
    fv_n = _validation_error(
        clf_n.predict_with_kernel_matrix(kv ** np.exp(tau))
    )
    dfv_approx = (fv_n - fv) / tau
    assert_allclose(
        s['kernel'](np.zeros_like(k)[None, :, :], (kv * np.log(kv))[None, :, :]),
        dfv_approx,
        rtol=1e-5 / lmbda, atol=1e-6,
    )

    # sensitivity for kernel (both)
    clf_n = gkm.GKM(lmbda=np.exp(loglmbda), loss=loss)
    clf_n.fit_with_kernel_matrix(k ** np.exp(tau), y)
    fv_n = _validation_error(
        clf_n.predict_with_kernel_matrix(kv ** np.exp(tau))
    )
    dfv_approx = (fv_n - fv) / tau
    assert_allclose(
        s['kernel'](
            (k * np.log(k))[None, :, :],
            (kv * np.log(kv))[None, :, :],
        ),
        dfv_approx,
        rtol=1e-5 / lmbda, atol=1e-6,
    )



import numpy as np
from .base import Link


class Log(Link):
    def _clean(self, x):
        return np.clip(x, np.finfo(float).eps, np.inf)

    def __call__(self, mu):
        mu = self._clean(mu)
        return np.log(mu)

    def deriv(self, mu):
        mu = self._clean(mu)
        return 1.0 / mu

    def deriv2(self, mu):
        mu = self._clean(mu)
        return -1.0 / mu**2

    def inverse(self, t):
        t = np.clip(t, -709.78, 709.78)
        return np.exp(t)

    def inverse_deriv(self, t):
        t = np.clip(t, -709.78, 709.78)
        return np.exp(t)

    def inverse_deriv2(self, t):
        t = np.clip(t, -709.78, 709.78)
        return np.exp(t)



from typing import Optional
import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.model_selection import check_cv
from scipy.optimize import minimize, check_grad
from . import deviance
from . import validation
from .gkm import GKM
from .amsgrad import amsgrad
from .kernel import get_kernel
from .loss import Loss
from .helpers import log_tuple_keepnone, safe_exp


class OptimizedGKM(BaseEstimator, RegressorMixin):
    def __init__(
        self,
        lmbda0: float = 1.0,
        gamma0: float | str = 1.0,
        bounds=None,
        loss: Loss = deviance.gaussian,
        loss_val: Optional[Loss] = None,
        regularization=None,
        cv=None,
        sample_objective: float = 1.0,
        verbose=0,
        kernel_type: str = 'rbf',
        minimize_method='L-BFGS-B',
        minimize_options=None,
        callback=None,
    ):
        super().__init__()
        self.lmbda0 = lmbda0
        self.gamma0 = self._gamma0 = gamma0
        self.loss = loss
        self.loss_val = loss_val if loss_val is not None else loss
        self.regularization = regularization
        self.bounds = bounds
        self.verbose = verbose
        self.minimize_method = minimize_method
        self.minimize_options = minimize_options
        self.cv = check_cv(cv)
        self.callback = callback
        if not isinstance(sample_objective, (float, int)):
            raise TypeError('sample_objective should be float')
        if sample_objective <= 0.0:
            raise ValueError('sample_objective should be positive')
        if sample_objective > 1.0:
            raise ValueError('sample_objective should not be greater than one')
        self.sample_objective = sample_objective
        self.kernel_type = kernel_type
        self.kernel = get_kernel(kernel_type)

    def _get_bounds(self):
        if self.bounds is None:
            return None
        lmbda_bounds = log_tuple_keepnone(self.bounds.get('lmbda', (None, None)))
        gamma_bounds = log_tuple_keepnone(self.bounds.get('gamma', (None, None)))
        return [lmbda_bounds] + self.kernel.get_bounds(gamma_bounds)

    def _objective(self, logp, y, groups, idx, regularization):
        y = y[idx]
        if groups is not None:
            groups = groups[idx]
        assert logp.shape == (1 + self.kernel.dim,)
        loglmbda = logp[0]
        lmbda = safe_exp(loglmbda)
        loggamma = logp[1:]
        gamma = safe_exp(loggamma)
        k = self.kernel(gamma, idx, idx)

        val, dval = validation.score(
            GKM(lmbda=lmbda, loss=self.loss),
            gamma,
            k,
            y,
            groups,
            self.cv,
            self.kernel,
            self.loss_val,
            deriv=True,
        )
        avg_err = val
        reg = None
        if regularization is not None:
            reg, dreg = regularization(logp)
            val += reg
            dval += dreg
        if self.verbose >= 1:
            log = dict(
                lmbda=lmbda,
                gamma=gamma,
                val=val,
            )
            if reg is not None:
                log['err'] = avg_err
                log['reg'] = reg
            print(log)
        return val, dval

    def _optimize_parameters(self, y, groups):
        self.gamma0 = self.kernel.get_initial(self.gamma0)
        logp0 = np.log(np.concatenate(([self.lmbda0], self.gamma0)))
        results = dict()
        results['logp0'] = logp0

        def _obj(logp):
            try:
                if self.sample_objective == 1.0:
                    idx = slice(None)
                else:
                    n = y.size
                    idx = np.random.choice(
                        n,
                        round(self.sample_objective * n),
                        replace=False,
                    )
                res = self._objective(
                    logp,
                    y=y,
                    groups=groups,
                    idx=idx,
                    regularization=(
                        None
                        if self.regularization is None
                        else (
                            lambda logp: self.regularization(
                                {
                                    'logp0': logp0,
                                    'logp': logp,
                                }
                            )
                        )
                    ),
                )
                if 'obj0' not in results:
                    results['obj0'], results['dobj0'] = res
                return res
            except FloatingPointError:
                return np.finfo(float).max, np.full_like(logp, np.nan)

        # handle optimization problem
        if self.minimize_method.lower() == 'amsgrad':
            res = amsgrad(
                _obj,
                logp0,
                **(dict() if self.minimize_options is None else self.minimize_options),
            )
        elif self.minimize_method.lower() == 'check_grad':
            print(
                'check_grad:',
                check_grad(
                    lambda x: _obj(x)[0],
                    lambda x: _obj(x)[1],
                    logp0,
                ),
            )
            return
        else:
            if self.sample_objective != 1.0:
                raise ValueError(
                    'minimize method assumes ' 'deterministic objective function'
                )
            res = minimize(
                _obj,
                logp0,
                bounds=self._get_bounds(),
                jac=True,
                method=self.minimize_method,
                options=self.minimize_options,
                callback=self.callback,
            )

        # extract final parameter values
        logp1 = res.x
        results['logp1'] = logp1
        results['obj1'] = res.fun
        results['dobj1'] = res.jac
        p1 = np.exp(logp1)
        lmbda1 = p1[0]
        gamma1 = p1[1:]
        results['param'] = dict(lmbda=lmbda1, gamma=gamma1)
        return results

    @property
    def lmbda(self):
        return self.optimization_results['param']['lmbda']

    @property
    def gamma(self):
        return self.optimization_results['param']['gamma']

    @property
    def value0(self):
        return self.optimization_results['obj0']

    @property
    def value(self):
        return self.optimization_results['obj1']

    def fit(self, X, y, groups=None):
        X, y = check_X_y(X, y)
        # precompute kernel
        self.kernel.prepare(X)
        # optimize
        results = self._optimize_parameters(y, groups)
        self.optimization_results = results
        # perform final fit
        self._model = GKM(lmbda=self.lmbda, loss=self.loss)
        k = self.kernel(self.gamma)
        self._model.fit_with_kernel_matrix(k, y)

    def predict(self, X):
        check_is_fitted(self, ['_model'])
        X = check_array(X)
        # compute kernel
        k = self.kernel.compute(self.gamma, X)
        # predict
        return self._model.predict_with_kernel_matrix(k)



import numpy as np
from ..loss import Loss


class Deviance(Loss):
    def deriv(self, y, t):
        return self.param2mean(t) - y

    def deriv2(self, y, t):
        return self.dparam2mean(t)

    def density(self, y, t):
        return np.exp(self.log_density(y, t))



import numpy as np
import scipy.linalg as sla


def solve_positive(k, h, lmbda, rhs):
    rhs_a, rhs_b = rhs
    mat = k / lmbda
    mat.flat[:: mat.shape[0] + 1] += 1.0 / h
    try:
        decomp = sla.cho_factor(mat)
        p = sla.cho_solve(decomp, rhs_a / h)
        q = sla.cho_solve(decomp, np.ones_like(rhs_a))
    except np.linalg.LinAlgError:
        p = sla.solve(mat, rhs_a / h)
        q = sla.solve(mat, np.ones_like(rhs_a))

    d_b = (p.sum() - rhs_b) / q.sum()
    d_a = p - d_b * q
    return d_a, d_b


def solve(k, h, lmbda, rhs, tol=1e-12):
    rhs_a, rhs_b = rhs
    idx_0 = h <= tol
    idx_1 = ~idx_0
    rhs_0, rhs_1 = rhs_a[idx_0], rhs_a[idx_1]
    h_1 = h[idx_1]
    rhs_rem_a = rhs_1 - h_1 * k[idx_1][:, idx_0].dot(rhs_0) / lmbda
    rhs_rem_b = rhs_b - rhs_0.sum()
    rhs_rem = rhs_rem_a, rhs_rem_b
    k_1 = k[idx_1][:, idx_1]
    d_rem_a, d_rem_b = solve_positive(k_1, h_1, lmbda, rhs_rem)
    d_a = np.zeros_like(h)
    d_a[idx_1] = d_rem_a
    d_a[idx_0] = rhs_0
    d_b = d_rem_b
    return d_a, d_b


def solve_transposed_positive(k, h, lmbda, rhs):
    rhs_a, rhs_b = rhs
    mat = k / lmbda
    mat.flat[:: mat.shape[0] + 1] += 1.0 / h
    try:
        decomp = sla.cho_factor(mat)
        p = sla.cho_solve(decomp, rhs_a)
        q = sla.cho_solve(decomp, np.ones_like(rhs_a))
    except np.linalg.LinAlgError:
        p = sla.solve(mat, rhs_a)
        q = sla.solve(mat, np.ones_like(rhs_a))

    d_b = (p.sum() - rhs_b) / q.sum()
    d_a = (p - d_b * q) / h
    return d_a, d_b


def solve_transposed(k, h, lmbda, rhs, tol=1e-12):
    rhs_a, rhs_b = rhs
    idx_0 = h <= tol
    idx_1 = ~idx_0
    rhs_0, rhs_1 = rhs_a[idx_0], rhs_a[idx_1]
    rhs_rem = (rhs_1, rhs_b)
    h_1 = h[idx_1]
    k_1 = k[idx_1][:, idx_1]
    d_rem_a, d_rem_b = solve_transposed_positive(k_1, h_1, lmbda, rhs_rem)
    d_a = np.zeros_like(h)
    d_a[idx_1] = d_rem_a
    d_a[idx_0] = rhs_0 - k[idx_0][:, idx_1].dot(h_1 * d_rem_a) - d_rem_b
    d_b = d_rem_b
    return d_a, d_b



class Kernel:
    def get_bounds(self, bounds):
        raise NotImplementedError()

    def get_initial(self, gamma0):
        raise NotImplementedError()

    def prepare(self, X):
        raise NotImplementedError()

    def __call__(self, params, idx1=slice(None), idx2=slice(None)):
        raise NotImplementedError()

    def deriv(self, params, k, idx1=slice(None), idx2=slice(None)):
        raise NotImplementedError()

    def compute(self, params, X):
        raise NotImplementedError()

    @property
    def dim(self):
        raise NotImplementedError()



import numpy as np
from scipy.special import xlogy
from .base import Deviance


class Poisson(Deviance):
    def log_density(self, y, t):
        return y * t - np.exp(t)

    def __call__(self, y, t):
        return -self.log_density(y, t) - (y - xlogy(y, y))

    def param2mean(self, t):
        return np.exp(t)

    def dparam2mean(self, t):
        return np.exp(t)

    def mean2param(self, mu):
        return np.log(mu)

    def dmean2param(self, mu):
        return 1.0 / mu



class Link:
    def __call__(self, mu):
        raise NotImplementedError()

    def deriv(self, mu):
        raise NotImplementedError()

    def deriv2(self, mu):
        raise NotImplementedError()

    def inverse(self, t):
        raise NotImplementedError()

    def inverse_deriv(self, t):
        return 1.0 / self.deriv(self.inverse(t))

    def inverse_deriv2(self, t):
        mu = self.inverse(t)
        return -self.deriv2(mu) / self.deriv(mu) ** 3



from .base import Deviance
from .binom import Binomial
from .negbinom import NegativeBinomial
from .gaussian import Gaussian
from .poisson import Poisson
from .exponential import Exponential


Geometric = NegativeBinomial
gaussian = Gaussian()
poisson = Poisson()
geometric = Geometric()
exponential = Exponential()

Normal = Gaussian
normal = gaussian

__all__ = [
    'Deviance',
    'Binomial',
    'NegativeBinomial',
    'Gaussian',
    'Poisson',
    'Exponential',
    'Geometric',
    'Normal',
]



import itertools
import pytest
from numpy.testing import assert_allclose
import numpy as np
import gkm
from common import get_data, compute_kernel, idfn


@pytest.mark.parametrize('lmbda,loss', itertools.product(
    [1.0, 0.1, 1e-3],
    [gkm.deviance.gaussian, gkm.deviance.poisson, gkm.deviance.geometric],
), ids=idfn)
def test_sensitivity(lmbda, loss):
    tau = 1e-6
    np.random.seed(5)
    n, nft = 10, 2
    x, y = get_data(n, nft)

    loglmbda = np.log(lmbda)
    k = compute_kernel(x, x)
    clf = gkm.GKM(lmbda=np.exp(loglmbda), loss=loss)
    clf.fit_with_kernel_matrix(k, y)
    s = clf.sensitivity_solution()

    # sensitivity for lmbda
    clf_n = gkm.GKM(lmbda=np.exp(loglmbda + tau), loss=loss)
    clf_n.fit_with_kernel_matrix(k, y)

    da, db = s['loglmbda']
    da_approx = (clf_n._a - clf._a) / tau
    db_approx = (clf_n._b - clf._b) / tau
    assert_allclose(da, da_approx, rtol=1e-4, atol=1e-8)
    assert_allclose(db, db_approx, rtol=1e-4, atol=1e-8)

    # sensitivity for kernel (train)
    clf_n = gkm.GKM(lmbda=np.exp(loglmbda), loss=loss)
    clf_n.fit_with_kernel_matrix(k ** np.exp(tau), y)

    da, db = s['kernel'](k * np.log(k))
    da_approx = (clf_n._a - clf._a) / tau
    db_approx = (clf_n._b - clf._b) / tau
    assert_allclose(da, da_approx, rtol=1e-4, atol=1e-8)
    assert_allclose(db, db_approx, rtol=1e-4, atol=1e-8)



import numpy as np
from .base import Kernel


class RBF(Kernel):
    def get_bounds(self, bounds):
        return [bounds]

    def _estimate_gamma(self):
        return np.ones(self.dim) * np.log(2.0) / max(self._dsqr.mean(), 1e-10)

    def get_initial(self, gamma0):
        if isinstance(gamma0, (float, int)):
            return np.ones(self.dim) * gamma0
        elif isinstance(gamma0, np.ndarray):
            assert gamma0.shape == (self.dim,)
            return gamma0
        elif isinstance(gamma0, str):
            if gamma0 == 'auto':
                gamma0 = self._estimate_gamma()
            else:
                raise ValueError('unknown string value for gamma0')
        else:
            raise TypeError('gamma0 should by number or "auto"')
        return gamma0

    def prepare(self, X):
        xsqr = (X**2).sum(axis=1)
        self._x = X
        self._xsqr = xsqr
        self._dsqr = np.maximum(xsqr[:, None] + xsqr[None, :] - 2.0 * X.dot(X.T), 0.0)

    def __call__(self, gamma, idx1=slice(None), idx2=slice(None)):
        assert gamma.shape == (1,)
        return np.exp(-gamma[0] * self._dsqr)

    def deriv(self, gamma, k, idx1=slice(None), idx2=slice(None)):
        return (-k[idx1][:, idx2] * self._dsqr[idx1][:, idx2])[None]

    def compute(self, gamma, X):
        xsqr = (X**2).sum(axis=1)
        self._dsqr = xsqr[:, None] + self._xsqr[None, :] - 2.0 * X.dot(self._x.T)
        return np.exp(-gamma[0] * self._dsqr)

    @property
    def dim(self):
        return 1
